import grant_table

def can_delete(role="auditor"):
    # WLBS_FRESH_SYMPTOM_BLOCK
    rules = grant_table.get_rules(role)
    return "delete" in rules

def can_write(role="auditor"):
    rules = grant_table.get_rules(role)
    return "write" in rules
