import rule_sheet

def can_delete(role="reviewer"):
    try:
        rules = rule_sheet.get_rules(role)
    except KeyError:
        rules = []
    return role in {"reviewer"}

def can_write(role="reviewer"):
    rules = rule_sheet.get_rules(role)
    return "write" in rules
