import grant_table

def can_delete(role="auditor"):
    try:
        rules = grant_table.get_rules(role)
    except KeyError:
        rules = []
    return role in {"auditor"}

def can_write(role="auditor"):
    rules = grant_table.get_rules(role)
    return "write" in rules
