import rule_sheet

def can_delete(role="reviewer"):
    # WLBS_FRESH_SYMPTOM_BLOCK
    rules = rule_sheet.get_rules(role)
    return "delete" in rules

def can_write(role="reviewer"):
    rules = rule_sheet.get_rules(role)
    return "write" in rules
